/*
 *	 NAME   : col.cpp
 *  AUTHOR : Arjun Krishna
 *  usage  : ./[prog_name] filename N 
 * "filename" contains data in edge List format
 *  N is the number of vertices in the Graph represented in "filename" 
 				{ Should be correct else leads to seg fault }
 */

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <list>
#include <map>

using namespace std;

// GLOBAL VARIABLES
int N; // N -- Number of Vertices --> argv[2]

ifstream edge;
//
class Graph
{
	// Parameters
	int V;			// no. of vertices
	list <int> * adj;		// adjacency List
	map <int,int> * xCol;	// A List to note what colors can't be used to color a particular vertex
	map <int,int> col;	//vertex number to color used map 
	int colors;	// number od colors used
	int * deg;	//array to store degree
	multimap <int,int> deg_ver; // for sorting degree
	int* ver_order;
public:
	//contructor
	Graph(int V,char* fname)	//Initialize Graph Parameters
	{
		this->V = V;
		adj = new list<int> [V];
		xCol = new map<int,int> [V];
		deg = (int*) malloc(sizeof(int)*V);
		deg_ver.clear();
		for(int i = 0;i<V;i++)
			deg[i] = 0;
		colors = 0;
		readGraph(fname);
		for(int i=0;i<V;i++)
			deg_ver.insert(make_pair(deg[i],i));
		ver_order = (int*) malloc(sizeof(int)*V);
	}
	void addEdge(int a,int b) //Add edge a,b to the Graph
	{
		adj[a].push_back(b);
		adj[b].push_back(a);
	}
	void readGraph(char* fname)	//Read Graph
	{
		edge.open(fname);
		int a,b;
		while(edge>>a>>b)
		{
			addEdge(a,b);
			(deg[a])++;
			(deg[b])++;
		}
		edge.close();
	}
	void rmVertex(int x,int c) //Remove vertex number x which shall be colored with c 
	{
		while(!adj[x].empty())
		{
			xCol[adj[x].back()].insert(make_pair(c,c));	//	add color 'c' to xcol map of all vertices adj to vertex x 
			adj[x].pop_back();	// delete vertex x
		}
		col[x] = c;			// add to map of colors;
	}
	int colMin(int x) // min Color needed to color vertex number x
	{
		map<int,int>::iterator itr;
		if(xCol[x].empty()) return 1;	//min color;
		else
		{
			int i = 1;
			for(itr = xCol[x].begin() ; itr!=xCol[x].end() ; itr++)
			{
				if((itr->first) != i)
					return i;
				i++;
			}
			return i;		// will return min color available
		}	
	}
	void vertexOrder()	
	{
	/* NORMAL ORDERING..without sorting reduces time but increases the number of colors used
	
		for(int i = 0;i<V;i++)		// using this will increase the number of colors used..
			ver_order[i] = i;				// if using this remove multiset and deg array..infact you can remove vertexOrder() itself
	
	______________________________________________________________________________ */
	/* USING SORTED DEGREE BASED ORDERING { vertices in decreasing order of degree } */
	
		multimap<int,int>::reverse_iterator itr;
		int i =0;
		for(itr = deg_ver.rbegin(); itr != deg_ver.rend() ; itr++)
			ver_order[i++] = itr->second;
	
	/*______________________________________________________________________________ */
	}
	void colorGraph()	// coloring algorithm
	{
		int c;
		vertexOrder();
		for(int i = 0 ; i<V ; i++)
		{
			c = colMin(ver_order[i]);		// identify the color to used to color ver_order[i]
			if( c > colors) colors = c;	
			rmVertex(ver_order[i],c);		// remove all the adjList members of ver_order[i] and see xcol list
		}		
	}
	void printColoring() //prints the coloring
	{
		map<int,int>::iterator itr;
		cout << "Colors Used : "<<colors <<endl;
		for(itr = col.begin(); itr!= col.end(); itr++)
			cout << itr->first << " " <<itr->second<<endl;
	}
};

/**MAIN**/

int main(int argc,char* argv[])
{
	if(argc!=3) { cerr<<"Usage : ./color fileName N "<<endl;return 0;}
	N = atoi(argv[2]);
	Graph G(N,argv[1]);
	G.colorGraph();
	G.printColoring();
	return 0;
}
