/****************************************************************************
Rahul Kejriwal

Greedy Coloring of graphs

Analysis: O(VE) for reading graphs
		  =>O(n^3) for cliques
		  O(n^2) for sorting
		  O(n^2) for coloring
****************************************************************************/

#include<iostream>
#include<vector>
#include<stdlib.h>
#include<fstream>

using namespace std;
using std::vector;

// vector<class node*> adjList;
class node** adjList;
int maxCol = 0;
int adjListSize;

class node{
	int val;
	int deg;
	int col;
	int degunsatur;
	vector<class node*> adj;

public:
	node(int num){
		val = num;
		deg = 0;
		col = -1;
		degunsatur = deg;
	}

	int getDegunsatur(){
		return degunsatur;
	}

	int getVal(){
		return val;
	}

	int getCol(){
		return col;
	}

	int getDeg(){
		return deg;
	}

	class node* getAdjI(int i){
		return adj[i];
	}

	void setCol(int color){
		col = color;
	}

	void update_degunsatur(){
		int i;
		for(i=0;i<adj.size();i++){
			(adj[i]->degunsatur)--;
		}
	}

	void addEdge(class node* adjNode){
		deg++;
		degunsatur++;
		adj.push_back(adjNode);
		adjNode->deg++;
		adjNode->degunsatur++;
		adjNode->adj.push_back(this);

	}

	void popEdge(){
		deg--;
		adj[adj.size()-1]->deg--;
		adj[adj.size()-1]->adj.pop_back();
		adj.pop_back();
	}

	void print_node(){
		cout<<"Vertex Val = "<<val<<",color = "<<col<<",deg = "<<deg;
		cout<<"\n";
	}
};

void print_graph(){
	int i,size = adjListSize;
	for(i=0; i<size; i++)
		adjList[i]->print_node();
}

void read_graph(char *fname){
	fstream graph(fname,ios::in);

	int n1,n2;

	int max = -1,tmp;
	while(graph){
		graph>>tmp;
		if(tmp>max){
			max = tmp;
		}
	}
	graph.close();

	adjListSize = max+1;

	graph.open(fname,ios::in);
	adjList = new class node*[max+1]{NULL};
	while(graph){
		graph>>n1>>n2;

		if(adjList[n1] == NULL){
			adjList[n1] = new class node(n1);
		}
		if(adjList[n2] == NULL){
			adjList[n2] = new class node(n2);
		}

		adjList[n1]->addEdge(adjList[n2]);
	}
	adjList[n1]->popEdge();
	graph.close();
}


void swap(int i,int j){
	class node *tmp = adjList[i];
	adjList[i] = adjList[j];
	adjList[j] = tmp;
}

int compare(int m,int n,int iter){
	if(iter%2)	
		return m<n;
	return m>n;
}

//Inclusive of minI and maxI element
void qsort_Col(int minI,int maxI,int iteration){
	
	if(minI >= maxI)
		return;

	int wall = minI,i;

	int curr = maxI;
	for(i=minI;i<maxI;i++){
		//compare call
		if(adjList[i]->getCol() > adjList[curr]->getCol() && iteration%2){
			swap(wall,i);
			++wall;
		}
		else if(adjList[i]->getCol() < adjList[curr]->getCol() && !(iteration%2)){
			swap(wall,i);
			++wall;	
		}
	}
	swap(wall,curr);
	if(wall>0)
		qsort_Col(minI,wall-1,iteration);
	if(wall<maxI)
		qsort_Col(wall+1,maxI,iteration);
}

class node* findMaxUnsatur(){
	int i;
	class node* max = NULL;
	int maxunsatur = 0;
	for(i=0;i<adjListSize;i++){
		if(adjList[i] == NULL)	continue;
		if(adjList[i]->getCol() != -1)
			continue;
		if(adjList[i]->getDegunsatur() > maxunsatur){
			max = adjList[i];
			maxunsatur = adjList[i]->getDegunsatur();
		}
	}
	return max;
}

void color_node(class node* vertex){
	
	if(vertex==NULL)
		return;

	int size = adjListSize;
	int *usedCol = new int[size];
	
	int i;
	for(i=0; i<size; i++)
		usedCol[i] = 0;

	for(i=0; i < vertex->getDeg(); i++){
		int currCol = vertex->getAdjI(i)->getCol(); 
		if(currCol != -1)
			usedCol[currCol] = 1;
	}

	i = 0;
	while(usedCol[i] == 1){
		i++;
	}
	vertex->setCol(i);	
	if(i > maxCol)
		maxCol = i;

	delete[] usedCol;
}

void dsatur(){
	int i;

	for(i=0;i<adjListSize;i++){
		class node* maxunsatur = findMaxUnsatur();
		if(maxunsatur){
			color_node(maxunsatur);
			maxunsatur->update_degunsatur();
		}
	}
}

void reset_Col(){
	int i;
	maxCol = 0;
	for(i=0; i<adjListSize; i++){
		adjList[i]->setCol(-1);
	}
}

void color_graph(){
	int i;
	for(i=0; i<adjListSize; i++){
		color_node(adjList[i]);
	}
}

void iter_greedy(int noOfIter){
	int i;
	for(i=0;i<noOfIter;i++){
		qsort_Col(0,adjListSize-1,i);
		reset_Col();
		color_graph();
	}
}

void gen_report(){
	fstream report("colDis.dat",ios::out);
	int i;
	int *dis = new int[maxCol+1]{0};

	for(i=0; i<adjListSize; i++){
		dis[adjList[i]->getCol()]++;
	}

	report<<"Color"<<"\t"<<"Frequency\n";
	for(i=0; i<maxCol; i++)
		report<<i<<"\t"<<dis[i]<<"\n";

	delete[] dis;
	report.close();
}

void gen_degReport(){
	fstream report("degDis.dat",ios::out);
	int i;
	int *dis = new int[adjListSize]{0};

	for(i=0; i<adjListSize; i++){
		dis[adjList[i]->getDeg()]++;
	}

	report<<"Degree"<<"\t"<<"Frequency\n";
	for(i=0; i<adjListSize; i++)
		if(dis[i] != 0)
			report<<i<<"\t"<<dis[i]<<"\n";

	delete[] dis;
	report.close();
}

int main(int argc, char* argv[]){
	read_graph(argv[1]);
	// print_graph();
	dsatur();
	iter_greedy(20);
	// cout<<"\n";
	// print_graph();
	gen_degReport();
	gen_report();

	cout<<"No of colors used: "<<maxCol+1<<"\nNo of nodes:"<<adjListSize<<"\n";

	return 0;
}