/*
 *	NAME   : hubcol.cpp
 *  AUTHOR : Arjun K
 *  usage  : ./[prog_name] filename N degRatio 
 * "filename" contains data in edge List format
 *  N is the number of vertices in the Graph represented in "filename"
 * (THRESHOLD) cutoff = (maxDegree) / DegRatio  //  degRatio  = (user_input)
 */

#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <list>
#include <map>
#include <vector>

using namespace std;

// GLOBAL VARIABLES
int N; // N -- Number of Vertices --> argv[2]

ifstream edge;
//
class Graph
{
	// Parameters
	int V;			// no. of vertices
	float degRatio;
	list <int> * adj;		// adjacency List
	map <int,int> * xCol;	// A List to note what colors can't be used to color a particular vertex
	map <int,int> col;	//vertex number to color used map 
	int colors;	// number od colors used
	int *deg;	//array to store degree
	multimap <int,int> deg_ver; // for sorting degree
	int* ver_order;
	int maxDeg;
	vector <int> hub;
	float cutoff;

public:
	//contructor
	Graph(int V,char* fname,float D)
	{
		maxDeg = 0;
		this->V = V;
		this->degRatio = D;
		adj = new list<int> [V];
		xCol = new map<int,int> [V];
		deg = (int*) malloc(sizeof(int)*V);
		for(int i = 0;i<V;i++)
			deg[i] = 0;
		colors = 0;
		degRead(fname);
		cutoff = (maxDeg/degRatio);	// Threshold or cutoff
		for(int i=0;i<V;i++)		// identifying hubs
		{
			if(deg[i] > cutoff) 
			{
				hub.push_back(i); 
				col[i] = colors+1;
				colors++;
			}
			else
			{
				deg_ver.insert(make_pair(deg[i],i));
			}
		}
		readGraph(fname);
		ver_order = (int*) malloc(sizeof(int)*(V-(hub.size())));
	}
	void degRead(char* fname)		//just read the degree
	{
		edge.open(fname);
		int a,b;
		while(edge>>a>>b)
		{
			deg[a]++;
			if(deg[a] >maxDeg) maxDeg = deg[a];
			deg[b]++;
			if(deg[b] >maxDeg) maxDeg = deg[b];
		}
		edge.close();
	}
	void addEdge(int a,int b) //Add edge a,b to the Graph
	{
		adj[a].push_back(b);
		adj[b].push_back(a);
	}
	int isHub(int ver)	// identify is ver is a hub
	{
		vector<int>::iterator itr;
		for(itr= hub.begin() ; itr!= hub.end();itr++)
			if(*itr == ver) return 1;
		return 0;
	}
	void readGraph(char* fname)	// adds Graph - hubs edges
	{ 
		edge.open(fname);
		int a,b,A,B;
		while(edge>>a>>b)
		{
			int flag = 0; //none is a hub
			A = isHub(a);
			B = isHub(b);
			if(A&&B) flag = 1;  //both A,B hubs
			if(flag == 0 && A) flag = 2;	// A hub
			if(flag == 0 && B) flag = 3;	// B hub
			switch(flag)
			{
				case 1: break;
			   case 0:  //none is a hub
			   {
			   	addEdge(a,b);
			   	break;	
			   }
			   case 2:	// a is a hub
			   {
			   	xCol[b].insert(make_pair(col[a],col[a]));
			   	break;
			   }
			   case 3:	// b is a hub
			   {
			   	xCol[a].insert(make_pair(col[b],col[b]));
			   	break;
			   }
			}
		
		}
		edge.close();
	}
	void rmVertex(int x,int c) //Remove vertex number x which shall be colored with c 
	{
		while(!adj[x].empty())
		{
			xCol[adj[x].back()].insert(make_pair(c,c));
			adj[x].pop_back();
		}
		col[x] = c;	
	}
	int colMin(int x) // min Color needed to color vertex number x
	{
		map<int,int>::iterator itr;
		if(xCol[x].empty()) return 1;
		else
		{
			int i = 1;
			for(itr = xCol[x].begin() ; itr!=xCol[x].end() ; itr++)
			{
				if((itr->first) != i)
					return i;
				i++;
			}
			return i;	
		}	
	}
	void vertexOrder()	
	{	
		multimap<int,int>::reverse_iterator itr;
		int i =0;
		for(itr = deg_ver.rbegin(); itr != deg_ver.rend() ; itr++)
			ver_order[i++] = itr->second;
	}
	void colorGraph()	// coloring algorithm
	{
		int c;
		vertexOrder();
		for(int i = 0 ; i<(V - (hub.size())) ; i++)
		{
			c = colMin(ver_order[i]);
			if( c > colors) colors = c;
			rmVertex(ver_order[i],c);
		}		
	}
	void printColoring() //prints the coloring
	{
		map<int,int>::iterator itr;
		cout << "Colors Used : "<<colors <<endl;
		for(itr = col.begin(); itr!= col.end(); itr++)
			cout << itr->first << " " <<itr->second<<endl;
	}
};

/**MAIN**/

int main(int argc,char* argv[])
{
	float D;
	if(argc!=4) { cerr<<"Usage : ./color fileName N degRatio"<<endl;return 0;}
	N = atoi(argv[2]);
	D = atof(argv[3]);
	Graph G(N,argv[1],D);
	G.colorGraph();
	G.printColoring();
	return 0;
}
